#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

#include "stream/stream.h"
#include "demuxer.h"
#include "stheader.h"

#include "muxer.h"
#include "aviheader.h"
#include "ms_hdr.h"
#include "mp_msg.h"
#include "help_mp.h"

#include "oggpcm.h"
#include "ogguvs.h"
struct ogg_stream_info {
    int initialized,eos,zeroct;
    double pts_offset;  //offset ensuring that at least one stream starts at 0
    oggpcm_stream_t pcm_stream;
    ogguvs_stream_t uvs_stream;
    FILE* file;
};

static void init_stream(muxer_stream_t *s){
  struct ogg_stream_info *si = s->priv;
  
  //look at the first packet to determine the starting pts
  //if it's less than the offset, or if the offset is unset
  if(s->muxer->muxbuf->pts < si->pts_offset 
      || si->pts_offset == MP_NOPTS_VALUE)
    //set the offset to this, the lowest starting pts among the streams
    si->pts_offset = s->muxer->muxbuf->pts;
  
  switch(s->type){
  case MUXER_TYPE_VIDEO: {
    int result;
    ogguvs_format_t fmt = OGGUVS_FMT_OTHER;
    ogguvs_colorsp_t csp = OGGUVS_CSP_UNSPEC_YCBCR;
    long img_size =0;

    mp_msg(MSGT_MUXER, MSGL_INFO, "[MUXER_OGG] Stream - VIDEO\n");
    mp_msg(MSGT_MUXER, MSGL_INFO, "[MUXER_OGG]   Size:        %d\n",
      s->bih->biSize);
    mp_msg(MSGT_MUXER, MSGL_INFO, "[MUXER_OGG]   Width:       %d\n",
      s->bih->biWidth);
    mp_msg(MSGT_MUXER, MSGL_INFO, "[MUXER_OGG]   Height:      %d\n",
      s->bih->biHeight);
    mp_msg(MSGT_MUXER, MSGL_INFO, "[MUXER_OGG]   Planes:      %d\n",
      s->bih->biPlanes);
    mp_msg(MSGT_MUXER, MSGL_INFO, "[MUXER_OGG]   BitCount:    %d\n",
      s->bih->biBitCount);
    mp_msg(MSGT_MUXER, MSGL_INFO, "[MUXER_OGG]   Compression: 0x%08X\n",
      s->bih->biCompression);
    mp_msg(MSGT_MUXER, MSGL_INFO, "[MUXER_OGG]   SizeImage:   %d\n",
      s->bih->biSizeImage);

    mp_msg(MSGT_MUXER, MSGL_V,
        "[MUXER_OGG] Writing stream header.\n");

    switch(s->bih->biCompression) {
    case 0x32315659:
        fmt = OGGUVS_FMT_YV12;
        img_size = s->bih->biWidth * s->bih->biHeight * 3/2;
        break;
    case 0x42475220:
        if(s->bih->biBitCount == 32) {
            /* No way to tell if this is ARGB or just RGB32 */
            fmt = OGGUVS_FMT_ARGBDIB;
            img_size = s->bih->biWidth * s->bih->biHeight * 4;
        }
        break;
    }

    result = (fmt != OGGUVS_FMT_OTHER);
    if(result)
        result = !si->initialized;
    else
        mp_msg(MSGT_MUXER, MSGL_WARN,
            "Video format not supported in this container, "
            "ignoring stream. (0x%08X)\n", s->bih->biCompression);

    {
        double fps=(float)s->h.dwRate/(float)s->h.dwScale;
        fps = (fps < 1000.0)?fps:0;

        if(result && ogguvs_stream_init(&si->uvs_stream,
                                         s->bih->biWidth,
                                         s->bih->biHeight,
                                         s->aspect,
                                         fps,
                                         90000,
                                         img_size,
                                         csp,
                                         fmt,
                                         s->ckid)) {
            mp_msg(MSGT_MUXER, MSGL_WARN,
                "OggUVS stream initialization failed, "
                "ignoring stream.\n");
            result = 0;
        }
    }
    if(result && ogguvs_header_pageout(&si->uvs_stream, si->file)) {
        mp_msg(MSGT_MUXER, MSGL_WARN,
                "Failed to write OggUVS header to output file, "
                "ignoring stream.\n");
        result = 0;
    }

    if(result)
        si->initialized = 1;
    break; }
  case MUXER_TYPE_AUDIO: {
      int result;
      mp_msg(MSGT_MUXER, MSGL_DBG2, "[MUXER_OGG] Stream - AUDIO\n");
      mp_msg(MSGT_MUXER, MSGL_DBG2, "[MUXER_OGG]   wFormatTag:      0x%08X\n",
        s->wf->wFormatTag);
      mp_msg(MSGT_MUXER, MSGL_DBG2, "[MUXER_OGG]   nChannels:       %d\n",
        s->wf->nChannels);
      mp_msg(MSGT_MUXER, MSGL_DBG2, "[MUXER_OGG]   nSamplesPerSec:  %d\n",
        s->wf->nSamplesPerSec);
      mp_msg(MSGT_MUXER, MSGL_DBG2, "[MUXER_OGG]   nAvgBytesPerSec: %d\n",
        s->wf->nAvgBytesPerSec);
      mp_msg(MSGT_MUXER, MSGL_DBG2, "[MUXER_OGG]   nBlockAlign:     %d\n",
        s->wf->nBlockAlign);
      mp_msg(MSGT_MUXER, MSGL_DBG2, "[MUXER_OGG]   wBitsPerSample:  %d\n",
        s->wf->wBitsPerSample);
      mp_msg(MSGT_MUXER, MSGL_DBG2, "[MUXER_OGG]   cbSize:          %d\n",
        s->wf->cbSize);

      mp_msg(MSGT_MUXER, MSGL_V,
          "[MUXER_OGG] Writing stream header.\n");

      result = (s->wf->wFormatTag == 1 && s->wf->wBitsPerSample == 16);
      if(result)
          result = !si->initialized;
      else
          mp_msg(MSGT_MUXER, MSGL_WARN,
              "Audio format not supported in this container, "
              "ignoring stream.\n");

      if(result && oggpcm_stream_init(&si->pcm_stream,
                                       s->wf->nChannels,
                                       s->wf->wBitsPerSample,
                                       s->wf->nSamplesPerSec,
                                       OGGPCM_FMT_S16_LE,
                                       s->ckid)) {
          mp_msg(MSGT_MUXER, MSGL_WARN,
              "OggPCM stream initialization failed, "
              "ignoring stream.\n");
          result = 0;
      }

      if(result && oggpcm_header_pageout(&si->pcm_stream, si->file)) {
          mp_msg(MSGT_MUXER, MSGL_WARN,
                  "Failed to write OggPCM header to output file, "
                  "ignoring stream.\n");
          result = 0;
      }

      if(result)
          si->initialized = 1;
    break; }
  }
}

static muxer_stream_t* oggfile_new_stream(muxer_t *muxer,int type){
    FILE* f;
    struct ogg_stream_info *si;
    muxer_stream_t* s;
    if(!muxer) return NULL;
    if(muxer->avih.dwStreams>=MUXER_MAX_STREAMS) {
	    mp_msg(MSGT_MUXER, MSGL_ERR, "Too many streams! increase MUXER_MAX_STREAMS !\n");
	    return NULL;
    }

    if(!(f= fdopen(muxer->stream->fd,"w"))) {
	    mp_msg(MSGT_MUXER, MSGL_ERR, "Unable to associate stream->fd w/a FILE* !\n");
	    return NULL;
    }
    /*ensure the stream is unbuffered*/
    setbuf(f,NULL);

    /* Initialze muxer stream info */
    s=malloc(sizeof(muxer_stream_t));
    if(!s) return NULL; // no mem!?
    memset(s,0,sizeof(muxer_stream_t));
    muxer->streams[muxer->avih.dwStreams]=s;
    s->type=type;
    s->id=muxer->avih.dwStreams;
    s->timer=0.0;
    s->size=0;
    s->muxer=muxer;
    s->priv=si=malloc(sizeof(struct ogg_stream_info));

    /* Initialize private data */
    memset(si,0,sizeof(struct ogg_stream_info));
    si->initialized = 0; si->file = f; si->pts_offset = MP_NOPTS_VALUE;

    switch(type){
    case MUXER_TYPE_VIDEO:
      mp_msg(MSGT_MUXER, MSGL_DBG2, "[MUXER_OGG] oggfile_new_stream MUXER_TYPE_VIDEO\n");
      s->ckid=mmioFOURCC(('0'+s->id/10),('0'+(s->id%10)),'d','c');
      s->h.fccType=streamtypeVIDEO;
      if(!muxer->def_v) muxer->def_v=s;
      break;
    case MUXER_TYPE_AUDIO:
      mp_msg(MSGT_MUXER, MSGL_DBG2, "[MUXER_OGG] oggfile_new_stream MUXER_TYPE_AUDIO\n");
      s->ckid=mmioFOURCC(('0'+s->id/10),('0'+(s->id%10)),'w','b');
      s->h.fccType=streamtypeAUDIO;
      break;
    default:
      mp_msg(MSGT_MUXER, MSGL_WARN, "Warning! unknown stream type: %d\n",type);
      return NULL;
    }
    muxer->avih.dwStreams++;
    return s;
}

static void oggfile_write_chunk(muxer_stream_t *s,size_t len,unsigned int flags, double dts, double pts){
    struct ogg_stream_info *si = s->priv;
    static int dbgframe=0;
       (void)dts;

    mp_msg(MSGT_MUXER, MSGL_DBG2, "[MUXER_OGG] oggfile_write_chunk %d - %.5lfs\n",
      dbgframe++, s->timer);

    if(!si->initialized) {
      init_stream(s);
      if(!si->initialized) return;
    }

    //offset the pts so that at least one stream starts at zero
    if (si->pts_offset != MP_NOPTS_VALUE) {
      pts -= si->pts_offset;
      if(pts<0.) {
        mp_msg(MSGT_MUXER, MSGL_WARN,
            "[MUXER_OGG] (%c) negative pts (%f/%f), discarding!\n",
            (s->type==MUXER_TYPE_AUDIO)?'A':'V',pts,si->pts_offset);
        return;
      }
    }
 
    // write out the chunk:
    switch(s->type) {
    case MUXER_TYPE_AUDIO:
        if(!len) {
          if(!si->eos&&++si->zeroct>25) {
            mp_msg(MSGT_MUXER, MSGL_WARN,
              "[MUXER_OGG] 0 byte audio threshold exceeded,"
              " creating artificial eos\n");
            oggpcm_footer_pageout(&si->pcm_stream, si->file);
            si->eos=1;
          }
          return;
        } else if(si->eos) {
          if(si->eos>0) {
            mp_msg(MSGT_MUXER, MSGL_WARN,
              "[MUXER_OGG] Warning: audio data rcv'd post artificial eos."
              " %d bytes dropped\n",len);
            si->eos=-1;
          }
          return;
        }

        if(!oggpcm_stream_packetin (&si->pcm_stream, s->buffer, len, pts)) {
            // Put the packet into the stream, now write out pages.
           //oggpcm_stream_pageout(&si->pcm_stream, si->file);
           oggpcm_stream_flush(&si->pcm_stream, si->file);
           si->zeroct=0;
        }
        break;
    case MUXER_TYPE_VIDEO:
        if(len < si->uvs_stream.field_size) {
            mp_msg(MSGT_MUXER, MSGL_V,
                "[MUXER_OGG] Dropping packet: len(%d) < field size(%d)\n",
                len, si->uvs_stream.field_size);
        } else if(len > si->uvs_stream.field_size) {
            mp_msg(MSGT_MUXER, MSGL_V,
                "[MUXER_OGG] Warning: len(%d) > field size(%d),"
                " using field size\n", len, si->uvs_stream.field_size);
            len = si->uvs_stream.field_size;
        }
        //Variable and fixed frame rates will be treated the same,
        // this enables a non-zero fps to be used (at worst it's avg fps)
        if(!ogguvs_stream_packetin (&si->uvs_stream, OGGUVS_FLD_TOP,
        s->buffer,len, pts)) {
            // Put the packet into the stream, now write out pages.
            //ogguvs_stream_pageout(&si->uvs_stream, si->file);
            ogguvs_stream_flush(&si->uvs_stream,si->file);
        }
        break;
    }
}

static void oggfile_write_header(muxer_t *muxer){
    unsigned int i;
    muxer_stream_t *s;
    struct ogg_stream_info *si;

    mp_msg(MSGT_MUXER, MSGL_DBG2, "[MUXER_OGG] oggfile_write_header\n");
    for(i=0;i<muxer->avih.dwStreams;i++){
        s = muxer->streams[i];
        si = s->priv;
        if(!si->initialized) {
          init_stream(s);
        } else {
          //make sure that we're at the EOF before writing the footer
          fseek(si->file, 0, SEEK_END);
            switch(s->type) {
            case MUXER_TYPE_VIDEO:
                mp_msg(MSGT_MUXER, MSGL_V,
                    "[MUXER_OGG] Writing stream footer.\n");
                if(ogguvs_footer_pageout(&si->uvs_stream, si->file)) {
                    mp_msg(MSGT_MUXER, MSGL_WARN,
                            "Failed to write OggUVS EOS packet.\n");
                }
                break;
            case MUXER_TYPE_AUDIO:
                mp_msg(MSGT_MUXER, MSGL_V,
                    "[MUXER_OGG] Writing stream footer.\n");
                if(oggpcm_footer_pageout(&si->pcm_stream, si->file)) {
                    mp_msg(MSGT_MUXER, MSGL_WARN,
                            "Failed to write OggPCM EOS packet.\n");
                }
                break;
            }
        }
    }
}

static void oggfile_write_index(muxer_t *muxer){
    mp_msg(MSGT_MUXER, MSGL_DBG2, "[MUXER_OGG] oggfile_write_index\n");
    muxer->cont_new_stream = &oggfile_new_stream;
       return;
}

int muxer_init_muxer_ogg(muxer_t *muxer){
    mp_msg(MSGT_MUXER, MSGL_DBG2, "[MUXER_OGG] Initializing Ogg Muxer.\n");
    muxer->cont_new_stream = &oggfile_new_stream;
    muxer->cont_write_chunk = &oggfile_write_chunk;
    muxer->cont_write_header = &oggfile_write_header;
    muxer->cont_write_index = &oggfile_write_index;
    return 1;
}
