/*******************************************************************************
*                                                                              *
* Copyright (c) 2005, On2 Technologies, Inc                                    *
* All rights reserved.                                                         *
*                                                                              *
* Redistribution and use in source and binary forms, with or without           *
* modification, are permitted provided that the following conditions are met:  *
*                                                                              *
*  - Redistributions of source code must retain the above copyright notice,    *
*    this list of conditions and the following disclaimer.                     *
*                                                                              *
*  - Redistributions in binary form must reproduce the above copyright notice, *
*    this list of conditions and the following disclaimer in the documentation *
*    and/or other materials provided with the distribution.                    *
*                                                                              *
*  - The names of the contributors may not be used to endorse or promote       *
*    products derived from this software without specific prior written        *
*    permission.                                                               *
*                                                                              *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"  *
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE    *
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE   *
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE     *
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR          *
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF         *
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS     *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN      *
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)      *
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE   *
* POSSIBILITY OF SUCH DAMAGE.                                                  *
*                                                                              *
*******************************************************************************/
#ifndef OGGPCM_H
#define OGGPCM_H
#include <stdio.h>
#include <ogg/ogg.h>

typedef enum {
    /* Integer coding */
    OGGPCM_FMT_S8       = 0x00000000, /* Signed integer 8 bit     */
    OGGPCM_FMT_U8       = 0x00000001, /* Unsigned integer 8 bit   */
    OGGPCM_FMT_S16_LE   = 0x00000002, /* Signed integer 16 bit little endian  */
    OGGPCM_FMT_S16_BE   = 0x00000003, /* Signed integer 16 bit big endian   */
    OGGPCM_FMT_S24_LE   = 0x00000004, /* Signed integer 24 bit little endian   */
    OGGPCM_FMT_S24_BE   = 0x00000005, /* Signed integer 24 bit big endian   */
    OGGPCM_FMT_S32_LE   = 0x00000006, /* Signed integer 32 bit little endian   */
    OGGPCM_FMT_S32_BE   = 0x00000007, /* Signed integer 32 bit big endian    */

    /* -- Compressed PCM */
    OGGPCM_FMT_ULAW     = 0x00000010, /* G.711 u-law encoding (8 bit)   */
    OGGPCM_FMT_ALAW     = 0x00000011, /* G.711 A-law encoding (8 bit)  */

    /* -- IEEE Floating point coding */
    OGGPCM_FMT_FLT32_LE = 0x00000020, /* IEEE Float [-1,1] 32 bit little endian  */
    OGGPCM_FMT_FLT32_BE = 0x00000021, /* IEEE Float [-1,1] 32 bit big endian     */
    OGGPCM_FMT_FLT64_LE = 0x00000022, /* IEEE Float [-1,1] 64 bit little endian   */
    OGGPCM_FMT_FLT64_BE = 0x00000023  /* IEEE Float [-1,1] 64 bit big endian */
} oggpcm_encoding_t;
                         
typedef struct {
    int               channels;
    int               sig_bits;
    int               frames_per_sec;
    int               bytes_per_frame;
    int               max_frames_per_pkt;
    oggpcm_encoding_t fmt;

    unsigned char     *pkt_data_payload;
    long              pkt_data_payload_sz;
    ogg_int64_t       packetno;
    ogg_int64_t       granulepos;

    /* Data below this point should be considered private. */
    int               num_extra_headers;
    int               num_extra_headers_read;
    long              pkt_alloc_size;
    ogg_packet        data_pkt;
    ogg_stream_state  ogg_ss;
} oggpcm_stream_t;

int oggpcm_stream_init (oggpcm_stream_t  *os,
                        int               channels,
                        int               sig_bits,
                        int               frames_per_sec,
                        oggpcm_encoding_t fmt,
                        int               serialno);
void oggpcm_stream_destroy(oggpcm_stream_t *os);
int oggpcm_stream_packetin (oggpcm_stream_t  *os, const void *buf, long size, double time);
int oggpcm_header_pageout (oggpcm_stream_t  *os, FILE *of);
int oggpcm_footer_pageout (oggpcm_stream_t  *os, FILE *of);
int oggpcm_stream_flush (oggpcm_stream_t  *os, FILE *of);
int oggpcm_stream_pageout (oggpcm_stream_t  *os, FILE *of);
int oggpcm_stream_parse_pkt (oggpcm_stream_t *os, ogg_packet *op);

#endif
