/*******************************************************************************
*                                                                              *
* Copyright (c) 2005, On2 Technologies, Inc                                    *
* All rights reserved.                                                         *
*                                                                              *
* Redistribution and use in source and binary forms, with or without           *
* modification, are permitted provided that the following conditions are met:  *
*                                                                              *
*  - Redistributions of source code must retain the above copyright notice,    *
*    this list of conditions and the following disclaimer.                     *
*                                                                              *
*  - Redistributions in binary form must reproduce the above copyright notice, *
*    this list of conditions and the following disclaimer in the documentation *
*    and/or other materials provided with the distribution.                    *
*                                                                              *
*  - The names of the contributors may not be used to endorse or promote       *
*    products derived from this software without specific prior written        *
*    permission.                                                               *
*                                                                              *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"  *
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE    *
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE   *
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE     *
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR          *
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF         *
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS     *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN      *
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)      *
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE   *
* POSSIBILITY OF SUCH DAMAGE.                                                  *
*                                                                              *
*******************************************************************************/
#ifndef OGGUVS_H
#define OGGUVS_H
#include <stdio.h>
#include <ogg/ogg.h>

#define OGGUVS_FMT_OTHER    0
#define OGGUVS_FMT_YV12     0x32315659L
#define OGGUVS_FMT_IYUV     0x56555949L
#define OGGUVS_FMT_YUY2     0x32595559L
#define OGGUVS_FMT_UYVY     0x59565955L
#define OGGUVS_FMT_YVYU     0x55595659L
#define OGGUVS_FMT_RGB24DIB 0x80808081L
#define OGGUVS_FMT_RGB32DIB 0x80808082L
#define OGGUVS_FMT_ARGBDIB  0x80808083L   
typedef unsigned long ogguvs_format_t;

typedef enum {
    OGGUVS_CSP_UNSPEC_RGB   = 0,
    OGGUVS_CSP_UNSPEC_YCBCR = 1
} ogguvs_colorsp_t;

typedef enum {
    OGGUVS_FLD_TOP    = 0,
    OGGUVS_FLD_BOTTOM = 1
} ogguvs_fieldid_t;

typedef struct {
    int               disp_width;
    int               disp_height;
    int               pix_aspect_num;
    int               pix_aspect_den;
    int               fld_rate_num;
    int               fld_rate_den;
    long              timebase;
    int               field_size;
    ogguvs_colorsp_t  colorspace;
    ogguvs_format_t   layout_id;
    int               interlaced;

    unsigned char     *pkt_data_payload;
    long              pkt_data_payload_sz;

    ogg_int64_t       packetno;
    ogg_int64_t       granulepos;

    /* Data below this point should be considered private. */
    unsigned char     *pkt_data;
    long              pkt_alloc_size;
    ogg_packet        data_pkt;
    ogg_stream_state  ogg_ss;
} ogguvs_stream_t;

int ogguvs_stream_init (ogguvs_stream_t  *os,
                        int               disp_width,
                        int               disp_height,
                        double            pixel_aspect,
                        double            field_rate,
                        long              timebase,
                        long              field_size,
                        ogguvs_colorsp_t  colorspace,
                        ogguvs_format_t   layout_id,
                        long              serialno);
void ogguvs_stream_destroy(ogguvs_stream_t  *os);

int ogguvs_stream_packetin (ogguvs_stream_t  *os,
                            ogguvs_fieldid_t  fld,
                            const void       *buf,
                            long              size,
                            double            time);

int ogguvs_header_pageout (ogguvs_stream_t  *os, FILE *of);
int ogguvs_footer_pageout (ogguvs_stream_t  *os, FILE *of);
int ogguvs_stream_flush (ogguvs_stream_t  *os, FILE *of);
int ogguvs_stream_pageout (ogguvs_stream_t  *os, FILE *of);
int ogguvs_stream_parse_pkt (ogguvs_stream_t *os, ogg_packet *op);

#endif
